@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 # coding=utf-8
-
+# from tee import StdoutTee, StderrTee
 import argparse
 import copy
 import getpass
@@ -19,10 +19,13 @@ import time
 import traceback
 from collections import OrderedDict
 from contextlib import contextmanager
+from typing import List, Optional
 
 import docker
 import termcolor
 import yaml
+from six import BytesIO
+
 from dt_shell.constants import DTShellConstants
 from dt_shell.env_checks import check_executable_exists, InvalidEnvironment, check_docker_environment
 from dt_shell.remote import ConnectionError, make_server_request, DEFAULT_DTSERVER
@@ -33,8 +36,6 @@ from duckietown_challenges.constants import CHALLENGE_SOLUTION_OUTPUT_DIR, CHALL
     CHALLENGE_DESCRIPTION_DIR, \
     CHALLENGE_EVALUATION_OUTPUT_DIR, ENV_CHALLENGE_NAME, ENV_CHALLENGE_STEP_NAME, CHALLENGE_PREVIOUS_STEPS_DIR
 from duckietown_challenges.utils import safe_yaml_dump, friendly_size, indent
-from six import BytesIO
-
 from . import __version__
 from .runner_cache import copy_to_cache, get_file_from_cache
 
@@ -75,8 +76,8 @@ def dt_challenges_evaluator():
         dt_challenges_evaluator_()
     except SystemExit:  # --help
         raise
-    except BaseException as e:
-        msg = traceback.format_exc(e)
+    except BaseException:
+        msg = traceback.format_exc()
         elogger.error(msg)
         sys.exit(2)
 
@@ -109,6 +110,7 @@ Usage:
     parser.add_argument("--reset", dest='reset', action="store_true", default=False,
                         help='Reset submission')
     parser.add_argument("--features", default='{}')
+    parser.add_argument("--debug-volumes", default=None)
     parsed = parser.parse_args(args=args)
 
     if parsed.version:
@@ -129,7 +131,7 @@ Usage:
     tmpdir = '/tmp/duckietown/DT18/evaluator/executions'
 
     try:
-        more_features = yaml.load(parsed.features)
+        more_features = yaml.load(parsed.features, Loader=yaml.SafeLoader)
     except BaseException as e:
         msg = 'Could not evaluate your YAML string %r:\n%s' % (parsed.features, e)
         raise Exception(msg)
@@ -149,7 +151,7 @@ Usage:
 
     args = dict(do_upload=do_upload, do_pull=do_pull, more_features=more_features,
                 delete=delete, evaluator_name=evaluator_name, machine_id=machine_id,
-                tmpdir=tmpdir, token=token)
+                tmpdir=tmpdir, token=token, debug_volumes=parsed.debug_volumes)
     ndots = 0
 
     if parsed.continuous:
@@ -177,12 +179,13 @@ Usage:
             except ConnectionError as e:
                 elogger.error(e)
                 multiplier *= multiplier_grow
-            except BaseException as e:
-                msg = 'Uncaught exception:\n%s' % traceback.format_exc(e)
+            except BaseException:
+                msg = 'Uncaught exception:\n%s' % traceback.format_exc()
                 elogger.error(msg)
                 multiplier *= multiplier_grow
 
-            time.sleep(timeout * multiplier)
+            r = random.random() * 2
+            time.sleep(timeout * multiplier + r)
 
     else:
         if parsed.submission:
@@ -270,11 +273,15 @@ def setup_logging(wd):
     old_stdout = sys.stdout
     if not os.path.exists(wd):
         os.makedirs(wd)
+    f_stderr = os.path.join(wd, 'stderr.log')
+    f_stdout = os.path.join(wd, 'stdout.log')
+
     my_stderr = Tee(os.path.join(wd, 'stderr.log'), sys.stderr)
     sys.stderr = my_stderr
     my_stdout = Tee(os.path.join(wd, 'stdout.log'), sys.stdout)
     sys.stdout = my_stdout
-    ch = logging.StreamHandler(my_stderr.file)
+    ch = logging.StreamHandler(sys.stderr.file)
+
     ch.setLevel(logging.DEBUG)
 
     root = logging.getLogger()
@@ -286,33 +293,53 @@ def setup_logging(wd):
     try:
         yield
     finally:
-
+        #
         my_stdout.flush()
         my_stderr.flush()
 
-        for s in [my_stderr, my_stdout]:
-            try:
-                fn = os.path.splitext(s.name)[0] + '.html'
-                with open(fn, 'w') as f:
-                    from ansi2html import Ansi2HTMLConverter
-                    conv = Ansi2HTMLConverter()
-                    v = s.getvalue().strip()
-                    if not v:
-                        v = '(No output was produced.)'
-                    html = conv.convert(v)
-                    f.write(html)
-                os.remove(s.name)
-            except:
-                pass
+        # for s in [my_stderr, my_stdout]:
+        #     try:
+        #         fn = os.path.splitext(s.name)[0] + '.html'
+        #         with open(fn, 'w') as f:
+        #             from ansi2html import Ansi2HTMLConverter
+        #             conv = Ansi2HTMLConverter()
+        #             v = s.getvalue().strip()
+        #             if not v:
+        #                 v = '(No output was produced.)'
+        #             html = conv.convert(v)
+        #             f.write(html)
+        #         os.remove(s.name)
+        #     except:
+        #         pass
 
         sys.stderr = old_stderr
         sys.stdout = old_stdout
         root.removeHandler(ch)
 
+    def convert(log):
+        fn = os.path.splitext(log)[0] + '.html'
+        data = open(log).read().strip()
+
+        from ansi2html import Ansi2HTMLConverter
+        conv = Ansi2HTMLConverter()
+        if not data:
+            data = '(No output was produced.)'
+        html = conv.convert(data)
+        with open(fn, 'w') as f:
+            f.write(html)
+        # os.remove(log)
+
+    convert(f_stdout)
+    convert(f_stderr)
+
 
 def go_(submission_id, do_pull, more_features, do_upload, delete, reset, evaluator_name, machine_id, tmpdir,
-        token=None):
+        token=None, debug_volumes: Optional[str] = None):
     features = get_features(more_features)
+    if features['processor_free_percent'] < 50:
+        elogger.info(f"Waiting because free CPU is {features['processor_free_percent']}%.")
+        return
+
     if token is None:
         token = get_token_from_shell_config()
     from duckietown_challenges import __version__ as version_dc
@@ -369,7 +396,7 @@ def go_(submission_id, do_pull, more_features, do_upload, delete, reset, evaluat
 
                 cr = run_single(wd, aws_config, steps2artefacts, challenge_parameters_, solution_container,
                                 challenge_name,
-                                challenge_step_name, project, do_pull)
+                                challenge_step_name, project, do_pull, debug_volumes=debug_volumes)
         finally:
             if do_upload:
                 uploaded = upload_files(wd, aws_config)
@@ -380,6 +407,9 @@ def go_(submission_id, do_pull, more_features, do_upload, delete, reset, evaluat
 
         if delete:
             shutil.rmtree(wd)
+        else:
+            msg = f'I will not delete temporary dir {wd}'
+            elogger.info(msg)
 
     except KeyboardInterrupt:
         msg = 'KeyboardInterrupt:\n%s' % traceback.format_exc()
@@ -426,7 +456,7 @@ def go_(submission_id, do_pull, more_features, do_upload, delete, reset, evaluat
 
 
 def run_single(wd, aws_config, steps2artefacts, challenge_parameters, solution_container, challenge_name,
-               challenge_step_name, project, do_pull):
+               challenge_step_name, project, do_pull, debug_volumes: Optional[str] = None):
     prepare_dir(wd, aws_config, steps2artefacts)
     tomonitor = []
     for service_name, service_def in challenge_parameters.services.items():
@@ -434,8 +464,11 @@ def run_single(wd, aws_config, steps2artefacts, challenge_parameters, solution_c
             continue
         else:
             tomonitor.append(service_name)
-    config = get_config(challenge_parameters, solution_container, challenge_name, challenge_step_name)
+    config = get_config(challenge_parameters, solution_container, challenge_name, challenge_step_name,
+                        debug_volumes=debug_volumes)
     config_yaml = yaml.safe_dump(config, encoding='utf-8', indent=4, allow_unicode=True)
+    if isinstance(config_yaml, bytes):
+        config_yaml = config_yaml.decode('utf-8')
     # elogger.debug('YAML:\n' + config_yaml)
 
     dcfn_original = os.path.join(wd, 'docker-compose.original.yaml')
@@ -450,14 +483,14 @@ def run_single(wd, aws_config, steps2artefacts, challenge_parameters, solution_c
     # validate the configuration
 
     try:
-        config_validated = run_docker(wd, project, ['config'], get_output=True)
+        config_validated = run_docker(wd, project, ['config'], get_output=True).decode('utf-8')
         with open(dcfn, 'w') as f:
             f.write(config_validated)
         elogger.info('config:\n%s' % config_validated)
         valid_config = True
         valid_config_error = None
-    except DockerComposeFail as e:
-        valid_config_error = 'Could not validate Docker Compose configuration:\n%s' % traceback.format_exc(e)
+    except DockerComposeFail:
+        valid_config_error = 'Could not validate Docker Compose configuration:\n%s' % traceback.format_exc()
         elogger.error(valid_config_error)
         valid_config = False
 
@@ -474,12 +507,12 @@ def run_single(wd, aws_config, steps2artefacts, challenge_parameters, solution_c
     return cr
 
 
-def get_id_for_service(wd, project, service_name):
+def get_id_for_service(wd, project, service_name) -> str:
     cmd = ['ps', '-q', service_name]
 
     try:
         o = run_docker(wd, project, cmd, get_output=True)
-        container_id = o.strip()  # \n at the end
+        container_id = o.decode('utf-8').strip()  # \n at the end
     except DockerComposeFail:
         raise
     return container_id
@@ -497,6 +530,7 @@ def write_logs_worker(service_name, container_id):
     container = client.containers.get(container_id)
     for x in container.logs(stream=True, stdout=True, stderr=True,
                             timestamps=True):
+        x = x.decode('utf-8')
         i = x.index(' ')
         timestamp = x[:i]
         message = x[i + 1:].rstrip()  # \n
@@ -535,8 +569,10 @@ def teminate_workers():
         p.terminate()
     elogger.debug('terminated all workers.')
 
+
 DEFAULT_TIMEOUT = 1800
 
+
 def run_one(wd, project, do_pull, monitor, services, timeout=DEFAULT_TIMEOUT):
     client = check_docker_environment()
 
@@ -569,19 +605,17 @@ def run_one(wd, project, do_pull, monitor, services, timeout=DEFAULT_TIMEOUT):
         for service_name in services:
             service2id[service_name] = get_id_for_service(wd, project, service_name)
             stream_logs(service_name, service2id[service_name])
-        # container = client.containers.get(container_id)
-
-        # last_notice = time.time()
         t0 = time.time()
 
         client = check_docker_environment()
         while True:
             exited = []
+            running = []
             should_break = False
             for service_name, container_id in service2id.items():
                 container = client.containers.get(container_id)
-                msg = 'service %s: %s' % (service_name, container.status)
-                elogger.debug(msg)
+                # msg = 'service %s: %s' % (service_name, container.status)
+                # elogger.debug(msg)
                 if container.status != 'running':
                     exited.append(service_name)
                     exit_code = exit_code_for_container(container_id)
@@ -589,7 +623,11 @@ def run_one(wd, project, do_pull, monitor, services, timeout=DEFAULT_TIMEOUT):
                         msg = '%s exited with code %s' % (service_name, exit_code)
                         elogger.error(msg)
                         should_break = True
+                else:
+                    running.append(service_name)
 
+            msg = 'Services running: %s  exited:  %s' % (running, exited)
+            elogger.debug(msg)
             if should_break:
                 break
 
@@ -663,7 +701,8 @@ def prepare_dir(wd, aws_config, steps2artefacts):
     download_artefacts(aws_config, steps2artefacts, previous_steps_dir)
 
 
-def get_config(challenge_parameters_, solution_container, challenge_name, challenge_step_name):
+def get_config(challenge_parameters_, solution_container, challenge_name, challenge_step_name,
+               debug_volumes: Optional[str] = None):
     client = docker.from_env()
     from docker.errors import NotFound
     # Adding the submission container
@@ -733,13 +772,19 @@ def get_config(challenge_parameters_, solution_container, challenge_name, challe
     # add volumes
 
     volumes = [
-        './' + CHALLENGE_SOLUTION_OUTPUT_DIR + ':' + '/' + CHALLENGE_SOLUTION_OUTPUT_DIR,
-        './' + CHALLENGE_RESULTS_DIR + ':' + '/' + CHALLENGE_RESULTS_DIR,
-        './' + CHALLENGE_DESCRIPTION_DIR + ':' + '/' + CHALLENGE_DESCRIPTION_DIR,
-        './' + CHALLENGE_EVALUATION_OUTPUT_DIR + ':' + '/' + CHALLENGE_EVALUATION_OUTPUT_DIR,
-        './' + CHALLENGE_PREVIOUS_STEPS_DIR + ':' + '/' + CHALLENGE_PREVIOUS_STEPS_DIR,
+        # './' + CHALLENGE_SOLUTION_OUTPUT_DIR + ':' + '/' + CHALLENGE_SOLUTION_OUTPUT_DIR,
+        # './' + CHALLENGE_RESULTS_DIR + ':' + '/' + CHALLENGE_RESULTS_DIR,
+        # './' + CHALLENGE_DESCRIPTION_DIR + ':' + '/' + CHALLENGE_DESCRIPTION_DIR,
+        # './' + CHALLENGE_EVALUATION_OUTPUT_DIR + ':' + '/' + CHALLENGE_EVALUATION_OUTPUT_DIR,
+        # './' + CHALLENGE_PREVIOUS_STEPS_DIR + ':' + '/' + CHALLENGE_PREVIOUS_STEPS_DIR,
+        '.:/challenges',
+        'fifos:/fifos'
     ]
 
+    if debug_volumes:
+        data = yaml.load(open(debug_volumes).read(), Loader=yaml.SafeLoader)
+        volumes.extend(data['volumes'])
+
     for service in config['services'].values():
         assert 'volumes' not in service
         service['volumes'] = copy.deepcopy(volumes)
@@ -751,6 +796,7 @@ def get_config(challenge_parameters_, solution_container, challenge_name, challe
     for service in config['services'].values():
         service['networks'] = copy.deepcopy(networks_evaluator)
     config['networks'] = dict(evaluation=None)
+    config['volumes'] = dict(fifos=None)
     return config
 
 
@@ -775,7 +821,7 @@ def write_logs(wd, project, services):
             logs = 'Service "%s" was not started.' % service
             elogger.warning(logs)
         else:
-            elogger.info('Found container ID = %r' % container_id)
+            # elogger.info('Found container ID = %r' % container_id)
 
             for fname, options in [
                 ('combined', dict(stderr=True, stdout=True)),
@@ -827,9 +873,9 @@ def write_logs(wd, project, services):
 #         return self.other.fileno()
 
 
-class Tee(object):
+class Tee:
     def __init__(self, name, f):
-        self.file = open(name, 'wb')
+        self.file = open(name, 'w')
         self.name = name
         self.original = f
 
@@ -842,9 +888,14 @@ class Tee(object):
     def __del__(self):
         self.close()
 
-    def write(self, data):
+    def write(self, data: str):
+        # if isinstance(data, str):
+        #     data = data.encode('utf-8')
         self.file.write(data)
+        self.file.flush()
+        # data_s = data.decode('utf-8')
         self.original.write(data)
+        self.original.flush()
 
     def flush(self):
         self.file.flush()
@@ -858,26 +909,27 @@ class Tee(object):
             return f.read()
 
 
-def run_docker(cwd, project, cmd0, get_output=False):
+def run_docker(cwd: str, project: str, cmd0: List[str], get_output=False) -> bytes:
     cmd0 = ['docker-compose', '-p', project] + cmd0
     # elogger.info('Running:\n\t%s' % " ".join(cmd0) + '\n\n in %s' % cwd)
-    elogger.debug('Running: %s' % " ".join(cmd0))
+    # elogger.debug('Running: %s' % " ".join(cmd0))
 
     d = tempfile.mkdtemp()
 
     fn1 = os.path.join(d, 'docker-stdout.txt')
     fn2 = os.path.join(d, 'docker-stderr.txt')
-    elogger.debug('Saving stdout to %s' % fn1)
+    # elogger.debug('Saving stdout to %s' % fn1)
 
+    # with StdoutTee(fn1), StderrTee(fn2):
 
     tee_stdout = Tee(fn1, sys.stdout)
     tee_stderr = Tee(fn2, sys.stderr)
 
-
     try:
         if get_output:
             return subprocess.check_output(cmd0, cwd=cwd, stderr=sys.stderr)
         else:
+            # noinspection PyTypeChecker
             subprocess.check_call(cmd0, cwd=cwd, stdout=tee_stdout, stderr=tee_stderr)
     except subprocess.CalledProcessError as e:
         msg = 'Could not run %s:\n\n %s' % (cmd0, indent(e, '  >  '))
@@ -893,7 +945,6 @@ def run_docker(cwd, project, cmd0, get_output=False):
         pass
 
 
-
 def upload_files(wd, aws_config, ignore_patterns=('.DS_Store',)):
     toupload = get_files_to_upload(wd, ignore_patterns=ignore_patterns)
 
@@ -934,7 +985,7 @@ def download_artefacts(aws_config, steps2artefacts, wd):
                 # no local
                 if 's3' in storage:
                     if not aws_config:
-                        msg = 'I cannot download from s3'
+                        msg = 'I cannot download from S3 because credentials not given.'
                         raise CouldNotDownloadAll(msg)
                     else:
                         s3ob = storage['s3']
@@ -1006,10 +1057,10 @@ def get_files_to_upload(path, ignore_patterns=()):
     return toupload
 
 
-def logs_for_container(client, container_id, stdout, stderr):
+def logs_for_container(client, container_id, stdout, stderr) -> str:
     container = client.containers.get(container_id)
     logs = container.logs(stdout=stdout, stderr=stderr, timestamps=True)
-    return logs
+    return logs.decode('utf-8')
 
 
 def only_copy_to_cache(toupload):
@@ -1097,10 +1148,11 @@ def object_exists(s3, bucket, key):
     return True
 
 
-def compute_sha256hex(filename):
+def compute_sha256hex(filename) -> str:
     cmd = ['shasum', '-a', '256', filename]
-    res = subprocess.check_output(cmd)
-    tokens = res.split()
+    res: bytes = subprocess.check_output(cmd)
+    res2 = res.decode('utf-8')
+    tokens = res2.split()
     h = tokens[0]
     assert len(h) == len('08c1fe03d3a6ef7dbfaccc04613ca561b11b5fd7e9d66b643436eb611dfba348')
     return h
