@@ -27,7 +27,7 @@ def get_file_from_cache(fn, sha256hex):
         raise KeyError(msg)
 
 
-def copy_to_cache(fn, sha256hex):
+def copy_to_cache(fn: str, sha256hex: str):
     if disable_cache:
         dclogger.warning('Forcing cache disabled.')
         return
