@@ -48,8 +48,8 @@ def read_submission_info(dirname):
     contents = open(fn).read()
     try:
         data = yaml.load(contents)
-    except BaseException as e:
-        raise CouldNotReadSubInfo(traceback.format_exc(e))
+    except BaseException:
+        raise CouldNotReadSubInfo(traceback.format_exc())
     try:
         return SubmissionDescription.from_yaml(data)
     except BaseException as e:
